package de.ozgcloud.archive.bescheid;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.stream.Stream;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedConstruction;
import org.mockito.Spy;

import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextAttachingInterceptor;
import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextProvider;
import de.ozgcloud.archive.common.callcontext.CallContextProvider;
import de.ozgcloud.archive.vorgang.VorgangWithEingangTestFactory;
import de.ozgcloud.document.bescheid.BescheidServiceGrpc.BescheidServiceBlockingStub;
import de.ozgcloud.document.bescheid.GrpcGetAllBescheidRequest;
import de.ozgcloud.document.bescheid.GrpcGetAllBescheidResponse;

class BescheidRemoteServiceTest {
	@Spy
	@InjectMocks
	private BescheidRemoteService service;
	@Mock
	private BescheidServiceBlockingStub grpcService;
	@Mock
	private BescheidMapper bescheidMapper;
	@Mock
	private CallContextProvider contextProvider;

	@Nested
	class TestFindByVorgangId {

		private final GrpcGetAllBescheidRequest request = GrpcGetAllBescheidRequestTestFactory.create();
		private final GrpcGetAllBescheidResponse response = GrpcGetAllBescheidResponseTestFactory.create();

		private final Bescheid bescheid = BescheidTestFactory.create();

		@Mock
		private BescheidServiceBlockingStub grpcServiceWithInterceptor;

		@BeforeEach
		void setUp() {
			doReturn(request).when(service).buildGetAllBescheidRequest(VorgangWithEingangTestFactory.ID);
			doReturn(grpcServiceWithInterceptor).when(service).getGrpcServiceWithInterceptor();
			when(grpcServiceWithInterceptor.getAll(any())).thenReturn(response);

		}

		@Test
		void shouldBuildGetAllRequest() {
			callService();

			verify(service).buildGetAllBescheidRequest(VorgangWithEingangTestFactory.ID);
		}

		@Test
		void shouldGetGrpcServiceWithInterceptor() {
			callService();

			verify(service).getGrpcServiceWithInterceptor();
		}

		@Test
		void shouldCallGrpcService() {
			callService();

			verify(grpcServiceWithInterceptor).getAll(request);
		}

		@Test
		void shouldMapBescheid() {
			callService().toList();

			verify(bescheidMapper).fromGrpc(GrpcGetAllBescheidResponseTestFactory.GRPC_BESCHEID, VorgangWithEingangTestFactory.ID);
		}

		@Test
		void shouldReturnStreamOfBescheide() {
			when(bescheidMapper.fromGrpc(any(), any())).thenReturn(bescheid);

			var bescheide = callService();

			assertThat(bescheide).containsExactly(bescheid);
		}

		private Stream<Bescheid> callService() {
			return service.findByVorgangId(VorgangWithEingangTestFactory.ID);
		}
	}

	@Nested
	class TestGetGrpcServiceWithInterceptor {

		private MockedConstruction<OzgCloudCallContextAttachingInterceptor> mockedConstructionInterceptor;
		private OzgCloudCallContextProvider argumentContextProvider;
		@Mock
		private OzgCloudCallContextAttachingInterceptor interceptor;

		@BeforeEach
		void setUpMocks() {
			mockedConstructionInterceptor = mockConstruction(OzgCloudCallContextAttachingInterceptor.class, (mock, context) -> {
				argumentContextProvider = (OzgCloudCallContextProvider) context.arguments().get(0);
				interceptor = mock;
			});
		}

		@AfterEach
		void cleanUp() {
			mockedConstructionInterceptor.close();
		}

		@Test
		void shoudCreateOzgCloudCallContextAttachingInterceptor() {
			callService();

			assertThat(mockedConstructionInterceptor.constructed()).hasSize(1);
		}

		@Test
		void shoudCreateOzgCloudCallContextAttachingInterceptorWithContextProvider() {
			callService();

			assertThat(argumentContextProvider).isEqualTo(contextProvider);
		}

		@Test
		void shouldSetInterceptor() {
			callService();

			verify(grpcService).withInterceptors(interceptor);
		}

		@Test
		void shouldReturnServiceStub() {
			when(grpcService.withInterceptors(any(OzgCloudCallContextAttachingInterceptor.class))).thenReturn(grpcService);

			var returnedServiceStub = callService();

			assertThat(returnedServiceStub).isEqualTo(grpcService);
		}

		private BescheidServiceBlockingStub callService() {
			return service.getGrpcServiceWithInterceptor();
		}
	}
}
