package de.ozgcloud.archive.common.command;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.Optional;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedConstruction;
import org.mockito.Spy;

import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextAttachingInterceptor;
import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextProvider;
import de.ozgcloud.archive.common.callcontext.CallContextProvider;
import de.ozgcloud.archive.vorgang.VorgangWithEingangTestFactory;
import de.ozgcloud.vorgang.grpc.command.CommandServiceGrpc.CommandServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.command.GrpcCommand;
import de.ozgcloud.vorgang.grpc.command.GrpcCommandsResponse;
import de.ozgcloud.vorgang.grpc.command.GrpcFindCommandsRequest;

class CommandRemoteServiceTest {

	@Spy
	@InjectMocks
	private CommandRemoteService service;
	@Mock
	private CommandServiceBlockingStub grpcService;
	@Mock
	private CommandMapper mapper;
	@Mock
	private CallContextProvider contextProvider;

	@Nested
	class TestFindCommands {

		private static final String COMMAND_STATUS_FINISHED = "FINISHED";
		private static final String REDIRECT_VORGANG_ORDER = "REDIRECT_VORGANG";
		private final GrpcCommand grpcCommand = GrpcCommand.newBuilder().build();
		private final GrpcCommandsResponse response = GrpcCommandsResponse.newBuilder().addAllCommand(Collections.singleton(grpcCommand)).build();

		@Mock
		private CommandServiceBlockingStub grpcServiceWithInterceptor;

		@Captor
		private ArgumentCaptor<GrpcFindCommandsRequest> requestCaptor;

		@BeforeEach
		void init() {
			doReturn(grpcServiceWithInterceptor).when(service).getGrpcServiceWithInterceptor();
			when(grpcServiceWithInterceptor.findCommands(any())).thenReturn(response);
		}

		@Test
		void shouldGetGrpcServiceWithInterceptor() {
			service.findCommands(VorgangWithEingangTestFactory.ID, Optional.empty(), Optional.empty());

			verify(service).getGrpcServiceWithInterceptor();
		}

		@Test
		void shouldCallGrpcService() {
			service.findCommands(VorgangWithEingangTestFactory.ID, Optional.empty(), Optional.empty());

			verify(grpcServiceWithInterceptor).findCommands(notNull());
		}

		@Test
		void shouldRequestWithVorgangId() {
			service.findCommands(VorgangWithEingangTestFactory.ID, Optional.empty(), Optional.empty());

			verify(grpcServiceWithInterceptor).findCommands(requestCaptor.capture());

			assertThat(requestCaptor.getValue().getVorgangId()).isEqualTo(VorgangWithEingangTestFactory.ID);
		}

		@Test
		void shouldRequestWithStatus() {
			service.findCommands(VorgangWithEingangTestFactory.ID, Optional.of(COMMAND_STATUS_FINISHED), Optional.empty());

			verify(grpcServiceWithInterceptor).findCommands(requestCaptor.capture());

			assertThat(requestCaptor.getValue().getStatusList()).contains(COMMAND_STATUS_FINISHED);
		}

		@Test
		void shouldRequestWithOrder() {
			service.findCommands(VorgangWithEingangTestFactory.ID, Optional.empty(), Optional.of(REDIRECT_VORGANG_ORDER));

			verify(grpcServiceWithInterceptor).findCommands(requestCaptor.capture());

			assertThat(requestCaptor.getValue().getOrderString()).isEqualTo(REDIRECT_VORGANG_ORDER);
		}

		@Test
		void shouldCallMapper() {
			service.findCommands(VorgangWithEingangTestFactory.ID, Optional.empty(), Optional.empty()).toList();

			verify(mapper).fromGrpc(grpcCommand);
		}
	}

	@Nested
	class TestGetGrpcServiceWithInterceptor {

		private MockedConstruction<OzgCloudCallContextAttachingInterceptor> mockedConstructionInterceptor;
		private OzgCloudCallContextProvider argumentContextProvider;
		@Mock
		private OzgCloudCallContextAttachingInterceptor interceptor;

		@BeforeEach
		void setUpMocks() {
			mockedConstructionInterceptor = mockConstruction(OzgCloudCallContextAttachingInterceptor.class, (mock, context) -> {
				argumentContextProvider = (OzgCloudCallContextProvider) context.arguments().get(0);
				interceptor = mock;
			});
		}

		@AfterEach
		void cleanUp() {
			mockedConstructionInterceptor.close();
		}

		@Test
		void shoudCreateOzgCloudCallContextAttachingInterceptor() {
			callService();

			assertThat(mockedConstructionInterceptor.constructed()).hasSize(1);
		}

		@Test
		void shoudCreateOzgCloudCallContextAttachingInterceptorWithContextProvider() {
			callService();

			assertThat(argumentContextProvider).isEqualTo(contextProvider);
		}

		@Test
		void shouldSetInterceptor() {
			callService();

			verify(grpcService).withInterceptors(interceptor);
		}

		@Test
		void shouldReturnServiceStub() {
			when(grpcService.withInterceptors(any(OzgCloudCallContextAttachingInterceptor.class))).thenReturn(grpcService);

			var returnedServiceStub = callService();

			assertThat(returnedServiceStub).isEqualTo(grpcService);
		}

		private CommandServiceBlockingStub callService() {
			return service.getGrpcServiceWithInterceptor();
		}
	}

}
