package de.ozgcloud.archive.export;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.fail;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.charset.StandardCharsets;
import java.util.Iterator;
import java.util.Optional;
import java.util.stream.Stream;
import java.util.zip.ZipInputStream;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.test.annotation.DirtiesContext;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.archive.bescheid.BescheidExportData;
import de.ozgcloud.archive.bescheid.ExportBescheidService;
import de.ozgcloud.archive.common.binaryfile.BinaryFileService;
import de.ozgcloud.archive.common.user.UserProfileTestFactory;
import de.ozgcloud.archive.common.user.UserService;
import de.ozgcloud.archive.file.ExportFileService;
import de.ozgcloud.archive.grpc.export.ExportServiceGrpc.ExportServiceBlockingStub;
import de.ozgcloud.archive.grpc.export.GrpcExportVorgangRequest;
import de.ozgcloud.archive.grpc.export.GrpcExportVorgangResponse;
import de.ozgcloud.archive.historie.ExportHistorieService;
import de.ozgcloud.archive.kommentar.ExportKommentarService;
import de.ozgcloud.archive.kommentar.KommentarTestFactory;
import de.ozgcloud.archive.kommentar.KommentarsExportData;
import de.ozgcloud.archive.kommentar.KommentarsExportDataTestFactory;
import de.ozgcloud.archive.postfach.ExportNachrichtService;
import de.ozgcloud.archive.postfach.PostfachMailExportData;
import de.ozgcloud.archive.vorgang.VorgangService;
import de.ozgcloud.archive.vorgang.VorgangTypeTestFactory;
import de.ozgcloud.archive.vorgang.VorgangWithEingang;
import de.ozgcloud.archive.vorgang.VorgangWithEingangTestFactory;
import de.ozgcloud.common.test.ITCase;
import lombok.SneakyThrows;
import net.devh.boot.grpc.client.inject.GrpcClient;

@SpringBootTest(properties = {
		"grpc.server.inProcessName=GrpcExportService-test",
		"grpc.client.inProcess.address=in-process:GrpcExportService-test"
})
@ITCase
@DirtiesContext
class ExportGrpcServiceITCase {

	@GrpcClient("inProcess")
	private ExportServiceBlockingStub serviceStub;

	@MockBean
	private ExportFileService exportFileService;
	@MockBean
	private VorgangService vorgangService;
	@MockBean
	private ExportHistorieService exportHistorieService;
	@MockBean
	private ExportKommentarService exportKommentarService;
	@MockBean
	private BinaryFileService binaryFileService;
	@MockBean
	private ExportNachrichtService exportNachrichtService;
	@MockBean
	private ExportBescheidService exportBescheidService;
	@MockBean
	private UserService userService;

	@Nested
	class TestExportVorgang {

		private final GrpcExportVorgangRequest request = GrpcExportVorgangRequestTestFactory.create();
		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

		@BeforeEach
		void setUpCommonMocks() {
			when(vorgangService.getVorgang(VorgangWithEingangTestFactory.ID)).thenReturn(vorgang);
			when(exportNachrichtService.createExportData(vorgang)).thenReturn(PostfachMailExportData.builder().build());
			when(exportBescheidService.createExportData(vorgang)).thenReturn(BescheidExportData.builder().build());
			when(exportFileService.getRepresentations(vorgang)).thenReturn(Stream.empty());
			when(exportFileService.getAttachments(vorgang)).thenReturn(Stream.empty());

			when(vorgangService.createKopf(vorgang)).thenReturn(NkAbgabeTypeTestFactory.create());
			when(vorgangService.createVorgangType(vorgang)).thenReturn(VorgangTypeTestFactory.create());
			when(vorgangService.createAkteType(vorgang)).thenReturn(AkteTypeTestFactory.create());

		}

		@Nested
		class OnMinimalVorgang {

			@BeforeEach
			void setUpMock() {
				when(exportKommentarService.createExportData(vorgang)).thenReturn(KommentarsExportData.builder().build());
			}

			@Test
			@SneakyThrows
			void shouldHaveOneZipEntry() {
				var response = exportVorgang();

				int numEntry;
				try (var outputStream = new ByteArrayOutputStream()) {
					collectResponseInStream(response, outputStream);
					numEntry = countZipEntries(outputStream);
				}

				assertThat(numEntry).isEqualTo(1);
			}

			@Test
			@SneakyThrows
			void shouldContainVorgangKopfDataInXml() {
				var response = exportVorgang();

				var content = new String[2];
				try (var outputStream = new ByteArrayOutputStream()) {
					collectResponseInStream(response, outputStream);
					writeZipEntryContentToString(outputStream, content);
				}

				assertThat(content[0]).isEqualToIgnoringWhitespace(VorgangWithEingangTestFactory.buildMinimalVorgangAsXml());
			}
		}

		@Nested
		class OnVorgangWithKommentarAndAttachment {
			private final String attachmentContent = LoremIpsum.getInstance().getWords(5);

			@BeforeEach
			void setUpMocks() {

				when(exportKommentarService.createExportData(vorgang)).thenReturn(KommentarsExportDataTestFactory.create());

				doAnswer(invocation -> {
					OutputStream outputStream = invocation.getArgument(1);
					outputStream.write(attachmentContent.getBytes());
					return null;
				}).when(exportFileService).writeOzgFile(any(), any());

				when(userService.getById(KommentarTestFactory.CREATED_BY)).thenReturn(Optional.of(UserProfileTestFactory.create()));
			}

			@Test
			@SneakyThrows
			void shouldHaveTwoZipEntries() {
				var response = exportVorgang();

				int numEntry;
				try (var outputStream = new ByteArrayOutputStream()) {
					collectResponseInStream(response, outputStream);
					numEntry = countZipEntries(outputStream);
				}

				assertThat(numEntry).isEqualTo(2);
			}

			@Test
			@SneakyThrows
			void shouldHaveKommentarAttachment() {
				var response = exportVorgang();

				String[] content = new String[2];
				try (var outputStream = new ByteArrayOutputStream()) {
					collectResponseInStream(response, outputStream);
					writeZipEntryContentToString(outputStream, content);
				}

				assertThat(content[1]).isEqualTo(attachmentContent);
			}
		}

		@SneakyThrows
		private void collectResponseInStream(Iterator<GrpcExportVorgangResponse> response, ByteArrayOutputStream outputStream) {
			while (response.hasNext()) {
				var next = response.next();
				if (StringUtils.EMPTY.equals(next.getVorgangFile().getFileName())) {
					try {
						outputStream.write(next.getVorgangFile().getFileContent().toByteArray());
					} catch (IOException e) {
						fail(e);
					}
				}
			}
		}

		@SneakyThrows
		private int countZipEntries(ByteArrayOutputStream outputStream) {
			var numEntry = 0;
			try (var inputStream = new ByteArrayInputStream(outputStream.toByteArray());
					var zipInputStream = new ZipInputStream(inputStream)) {
				while ((zipInputStream.getNextEntry()) != null) {
					numEntry++;
				}
			}
			return numEntry;
		}

		@SneakyThrows
		private void writeZipEntryContentToString(ByteArrayOutputStream outputStream, String[] content) {
			try (var byteArrayInputStream = new ByteArrayInputStream(outputStream.toByteArray());
					var zipInputStream = new ZipInputStream(byteArrayInputStream)) {
				var numEntry = 0;
				while ((zipInputStream.getNextEntry()) != null) {
					content[numEntry] = IOUtils.toString(zipInputStream, StandardCharsets.UTF_8);
					numEntry++;
				}
			}
		}

		private Iterator<GrpcExportVorgangResponse> exportVorgang() {
			return serviceStub.exportVorgang(request);
		}
	}
}