package de.ozgcloud.archive.export;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.OutputStream;
import java.util.UUID;
import java.util.function.Consumer;
import java.util.function.Function;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedConstruction;
import org.mockito.Spy;
import org.springframework.core.task.TaskExecutor;

import com.google.protobuf.ByteString;
import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.archive.grpc.export.GrpcExportVorgangRequest;
import de.ozgcloud.archive.grpc.export.GrpcExportVorgangResponse;
import de.ozgcloud.archive.grpc.export.GrpcFile;
import de.ozgcloud.archive.vorgang.VorgangWithEingangTestFactory;
import de.ozgcloud.common.binaryfile.GrpcBinaryFileServerDownloader;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;

class ExportGrpcServiceTest {

	@InjectMocks
	@Spy
	private ExportGrpcService service;
	@Mock
	private ExportService exportService;
	@Mock
	private TaskExecutor taskExecutor;

	@Nested
	class TestExportVorgang {

		@Mock
		private StreamObserver<GrpcExportVorgangResponse> responseObserver;
		@Mock
		private GrpcBinaryFileServerDownloader<GrpcExportVorgangResponse> downloader;

		private final GrpcExportVorgangRequest request = GrpcExportVorgangRequestTestFactory.create();
		private final String fileNameId = UUID.randomUUID().toString();

		@BeforeEach
		void setUpMocks() {
			doReturn(fileNameId).when(service).createFileNameId();
			doReturn(downloader).when(service).buildExportDownloader(VorgangWithEingangTestFactory.ID, fileNameId, responseObserver);
		}

		@Test
		void shouldCreateFileNameId() {
			callService();

			verify(service).createFileNameId();
		}

		@Test
		void shouldSendFileName() {
			callService();

			verify(service).sendFileName(responseObserver, fileNameId);
		}

		@Test
		void shouldBuildExportDownloader() {
			callService();

			verify(service).buildExportDownloader(VorgangWithEingangTestFactory.ID, fileNameId, responseObserver);
		}

		@Test
		void shouldStartDownloader() {
			callService();

			verify(downloader).start();
		}

		private void callService() {
			service.exportVorgang(request, responseObserver);
		}
	}

	@Nested
	class TestCreateFileNameId {

		private final UUID uuid = UUID.randomUUID();

		@Test
		void shouldReturnRandomUUID() {
			try (var mockedStaticUUID = mockStatic(UUID.class)) {
				mockedStaticUUID.when(UUID::randomUUID).thenReturn(uuid);

				var returnedFileNameId = service.createFileNameId();

				assertThat(returnedFileNameId).isEqualTo(uuid.toString());
			}
		}
	}

	@Nested
	class TestSendFileName {

		@Mock
		private StreamObserver<GrpcExportVorgangResponse> responseObserver;
		@Mock
		private GrpcBinaryFileServerDownloader<GrpcExportVorgangResponse> downloader;

		private final String fileNameId = UUID.randomUUID().toString();
		private final String fileName = LoremIpsum.getInstance().getName();

		@Test
		void shouldCallBuildXdomeaFileName() {
			sendFileName();

			verify(service).buildXdomeaFileName(fileNameId);
		}

		@Test
		void shouldSendFileName() {
			doReturn(fileName).when(service).buildXdomeaFileName(fileNameId);

			sendFileName();

			verify(responseObserver).onNext(argThat((response) -> response.getVorgangFile().getFileName().equals(fileName)));
		}

		private void sendFileName() {
			service.sendFileName(responseObserver, fileNameId);
		}
	}

	@Nested
	class TestBuildExportDownloader {

		@Mock
		private CallStreamObserver<GrpcExportVorgangResponse> responseObserver;

		private final String fileNameId = UUID.randomUUID().toString();
		@Mock
		private OutputStream outputStream;

		private MockedConstruction<GrpcBinaryFileServerDownloader> downloaderMockedConstruction; // NOSONAR
		private GrpcBinaryFileServerDownloader<GrpcExportVorgangResponse> downloader;
		private StreamObserver<GrpcExportVorgangResponse> setResponseObserver;
		private TaskExecutor setTaskExecutor;
		private Consumer<OutputStream> setDownloadConsumer;
		private Function<ByteString, GrpcExportVorgangResponse> setChunkBuilder;

		private final ByteString chunk = ByteString.copyFromUtf8(LoremIpsum.getInstance().getWords(5));

		@SuppressWarnings("unchecked")
		@BeforeEach
		void mock() {
			downloaderMockedConstruction = mockConstruction(GrpcBinaryFileServerDownloader.class, (downloader, context) -> {
				setResponseObserver = (StreamObserver<GrpcExportVorgangResponse>) context.arguments().get(0);
				setChunkBuilder = (Function<ByteString, GrpcExportVorgangResponse>) context.arguments().get(1);
				setDownloadConsumer = (Consumer<OutputStream>) context.arguments().get(2);
				setTaskExecutor = (TaskExecutor) context.arguments().get(3);
				this.downloader = downloader;
			});
		}

		@AfterEach
		void closeMock() {
			downloaderMockedConstruction.close();
		}

		@Test
		void shouldSetResponseObserver() {
			buildExportDownloader();

			assertThat(setResponseObserver).isEqualTo(responseObserver);
		}

		@Test
		void shouldSetTaskExecutor() {
			buildExportDownloader();

			assertThat(setTaskExecutor).isEqualTo(taskExecutor);
		}

		@Test
		void shouldSetDownloadConsumer() {
			buildExportDownloader();

			setDownloadConsumer.accept(outputStream);

			verify(exportService).writeXdomeaFileContent(VorgangWithEingangTestFactory.ID, fileNameId, outputStream);
		}

		@Test
		void shouldSetChunkBuilder() {
			var response = GrpcExportVorgangResponse.newBuilder().setVorgangFile(GrpcFile.newBuilder().setFileContent(chunk).build()).build();
			doReturn(response).when(service).buildExportVorgangChunkResponse(chunk);
			buildExportDownloader();

			var result = setChunkBuilder.apply(chunk);

			assertThat(result).isEqualTo(response);
		}

		@Test
		void shouldReturnDownloader() {
			var returnedDownloader = buildExportDownloader();

			assertThat(returnedDownloader).isEqualTo(downloader);
		}

		private GrpcBinaryFileServerDownloader<GrpcExportVorgangResponse> buildExportDownloader() {
			return service.buildExportDownloader(VorgangWithEingangTestFactory.ID, fileNameId, responseObserver);
		}
	}

	@Nested
	class TestBuildExportVorgangChunkResponse {

		private final ByteString chunk = ByteString.copyFromUtf8(LoremIpsum.getInstance().getWords(5));

		@Test
		void shouldReturnResponse() {
			var expectedResponse = GrpcExportVorgangResponse.newBuilder().setVorgangFile(GrpcFile.newBuilder().setFileContent(chunk).build()).build();

			var result = service.buildExportVorgangChunkResponse(chunk);

			assertThat(result).isEqualTo(expectedResponse);

		}
	}

	@Nested
	class TestBuildXdomeaFileName {

		private final String id = UUID.randomUUID().toString();

		@Test
		void shouldReturnFileName() {
			var fileName = service.buildXdomeaFileName(id);

			assertThat(fileName).isEqualTo(id + "_Abgabe.Abgabe.0401.xdomea");
		}
	}
}
