package de.ozgcloud.archive.historie;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.mockito.Mockito;
import org.mockito.Spy;
import org.testcontainers.shaded.org.apache.commons.lang3.NotImplementedException;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.archive.common.command.ArchiveManagerCommand;
import de.ozgcloud.archive.common.command.CommandTestFactory;
import de.ozgcloud.archive.common.command.CommandWithPrevious;
import de.ozgcloud.archive.common.command.CommandWithPreviousTestFactory;
import de.ozgcloud.archive.common.user.UserProfileTestFactory;
import de.ozgcloud.archive.historie.ChangeHistoryBuilder.CommandWithChangeValues;

public class ChangeHistoryBuilderTest {

	private static final String ORGANISATIONSEINHEITEN_ID = UUID.randomUUID().toString();

	private final ArchiveManagerCommand previousCommand = commandFinishedAt(LocalDateTime.of(2023, 5, 1, 12, 0));
	private final ArchiveManagerCommand command = commandFinishedAt(LocalDateTime.of(2023, 6, 1, 12, 0));
	private final List<ArchiveManagerCommand> commands = List.of(previousCommand, command);

	@Spy
	private TestChangeHistoryBuilder builder = new TestChangeHistoryBuilder()
			.withCommands(commands)
			.withOrganisationseinheitenID(ORGANISATIONSEINHEITEN_ID);

	@Nested
	class TestBuild {

		private final CommandWithChangeValues previousCommandWithChangeValues = new CommandWithChangeValues(
				previousCommand, "a", "b");
		private final CommandWithChangeValues commandWithChangeValues = new CommandWithChangeValues(
				command, "b", "c");
		private final List<CommandWithChangeValues> commandsWithChangeValues = List.of(previousCommandWithChangeValues,
				commandWithChangeValues);

		@Test
		void shouldCallInOrder() {
			var orderVerifier = Mockito.inOrder(builder);
			doReturn(commands.stream()).when(builder).retrieveRelevantCommands();
			doReturn(commandsWithChangeValues.stream()).when(builder).toCommandsWithChangeValues(notNull());

			builder.build().toList();

			orderVerifier.verify(builder).retrieveRelevantCommands();
			orderVerifier.verify(builder).inChronologicalOrder(notNull());
			orderVerifier.verify(builder).addPreviousCommand(notNull());
			orderVerifier.verify(builder).toCommandsWithChangeValues(notNull());
			orderVerifier.verify(builder).toVorgangChanges(notNull());
		}
	}

	@Nested
	class TestRetrieveRelevantCommands {

		@BeforeEach
		void init() {
			doReturn(false).when(builder).isRelevant(previousCommand);
			doReturn(true).when(builder).isRelevant(command);
		}

		@Test
		void shouldCallIsRelevant() {
			builder.retrieveRelevantCommands().toList();

			verify(builder).isRelevant(previousCommand);
			verify(builder).isRelevant(command);
		}

		@Test
		void shouldFilterOutIrrelevantCommands() {
			var result = builder.retrieveRelevantCommands().toList();

			assertThat(result).containsExactly(command);
		}
	}

	@Nested
	class TestInChronologicalOrder {

		private final Stream<ArchiveManagerCommand> unorderedCommands = Stream.of(command, previousCommand);

		@Test
		void shouldSortAscendingByFinishedAt() {
			var orderedCommands = builder.inChronologicalOrder(unorderedCommands);

			assertThat(orderedCommands).containsExactly(previousCommand, command);
		}
	}

	@Nested
	class TestAddPreviousCommand {

		@Test
		void shouldSetPreviousOfFirstCommandToNull() {
			var commands = Stream.of(previousCommand);

			var commandsWithPrevious = builder.addPreviousCommand(commands);

			assertThat(commandsWithPrevious).first().extracting("previous").isNull();
		}

		@Test
		void shouldSetPreviousOfSubsequentCommand() {
			var commands = Stream.of(previousCommand, command);

			var commandsWithPrevious = builder.addPreviousCommand(commands);

			assertThat(commandsWithPrevious).element(1).extracting("previous").isEqualTo(previousCommand);
		}
	}

	@Nested
	class TestToCommandsWithChangeValues {

		private final CommandWithPrevious commandWithoutPrevious = CommandWithPreviousTestFactory.createBuilder().command(previousCommand)
				.previous(null).build();
		private final CommandWithPrevious commandWithPrevious = CommandWithPreviousTestFactory.createBuilder().command(command)
				.previous(previousCommand).build();
		private final List<CommandWithPrevious> commandsWithPrevious = List.of(commandWithoutPrevious, commandWithPrevious);

		private final ChangeHistoryBuilder.CommandWithChangeValues previousCommandWithChangeValues = new ChangeHistoryBuilder.CommandWithChangeValues(
				previousCommand, "a", "b");
		private final ChangeHistoryBuilder.CommandWithChangeValues commandWithChangeValues = new ChangeHistoryBuilder.CommandWithChangeValues(
				command, "b", "c");

		@BeforeEach
		void init() {
			doReturn(previousCommandWithChangeValues).when(builder).toCommandWithChangeValues(commandWithoutPrevious);
			doReturn(commandWithChangeValues).when(builder).toCommandWithChangeValues(commandWithPrevious);
		}

		@Test
		void shouldMapToCommandWithChangeValues() {
			callBuilder().toList();

			verify(builder).toCommandWithChangeValues(commandWithoutPrevious);
			verify(builder).toCommandWithChangeValues(commandWithPrevious);
		}

		@Test
		void shouldReturnMappedCommandWithChangeValues() {
			var historieEntries = callBuilder().toList();

			assertThat(historieEntries).containsExactly(previousCommandWithChangeValues, commandWithChangeValues);
		}

		private Stream<CommandWithChangeValues> callBuilder() {
			return builder.toCommandsWithChangeValues(commandsWithPrevious.stream());
		}
	}

	@Nested
	class TestToVorgangChange {

		private final ChangeHistoryBuilder.CommandWithChangeValues commandWithChangeValues = new ChangeHistoryBuilder.CommandWithChangeValues(
				previousCommand, "a", "b");

		@Test
		void shouldSeValueBeforeChange() {
			var vorgangChange = callBuilder();

			assertThat(vorgangChange.getValueBeforeChange()).isNotBlank().isEqualTo(commandWithChangeValues.valueBeforeChange());
		}

		@Test
		void shouldSeValueAfterChange() {
			var vorgangChange = callBuilder();

			assertThat(vorgangChange.getValueAfterChange()).isNotBlank().isEqualTo(commandWithChangeValues.valueAfterChange());
		}

		@Test
		void shouldSetAuthorFullName() {
			var vorgangChange = callBuilder();

			assertThat(vorgangChange.getAuthorFullName()).isEqualTo(UserProfileTestFactory.FULLNAME);
		}

		@Test
		void shouldSetFinishedAt() {
			var vorgangChange = callBuilder();

			assertThat(vorgangChange.getFinishedAt()).isNotNull().isEqualTo(previousCommand.getFinishedAt());
		}

		@Test
		void shouldSetOrder() {
			var vorgangChange = callBuilder();

			assertThat(vorgangChange.getOrder()).isNotNull().isEqualTo(previousCommand.getOrder());
		}

		@Test
		void shouldSetOrganisationseinheitenID() {
			var vorgangChange = callBuilder();

			assertThat(vorgangChange.getOrganisationseinheitenID()).isEqualTo(ORGANISATIONSEINHEITEN_ID);
		}

		private VorgangChange callBuilder() {
			return builder.toVorgangChange(commandWithChangeValues);
		}
	}

	@Nested
	class TestGetValueFromCommandBody {

		private static final String PROPERTY_NAME = LoremIpsum.getInstance().getWords(1);

		@ParameterizedTest
		@NullAndEmptySource
		void shouldReturnEmptyIfBodyIsNullOrEmpty(Map<String, ?> body) {
			var command = previousCommand.toBuilder().body(body).build();

			var value = builder.getValueFromCommandBody(PROPERTY_NAME, command);

			assertThat(value).isEmpty();
		}

		@Test
		void shouldReturnEmptyIfPropertyIsNotPresentInBody() {
			var command = previousCommand.toBuilder().body(Map.of("a", "b")).build();

			var value = builder.getValueFromCommandBody(PROPERTY_NAME, command);

			assertThat(value).isEmpty();
		}

		@Test
		void shouldReturnValueFromBody() {
			var expectedValue = LoremIpsum.getInstance().getWords(2);
			var command = previousCommand.toBuilder().body(Map.of(PROPERTY_NAME, expectedValue)).build();

			var value = builder.getValueFromCommandBody(PROPERTY_NAME, command);

			assertThat(value).isEqualTo(expectedValue);
		}
	}

	static ArchiveManagerCommand commandFinishedAt(LocalDateTime finishedAt) {
		return CommandTestFactory.createBuilder()
				.finishedAt(ZonedDateTime.of(finishedAt, ZoneId.of("UTC")))
				.build();
	}

	private static class TestChangeHistoryBuilder extends ChangeHistoryBuilder<TestChangeHistoryBuilder> {

		@Override
		boolean isRelevant(ArchiveManagerCommand command) {
			throw new NotImplementedException();
		}

		@Override
		CommandWithChangeValues toCommandWithChangeValues(CommandWithPrevious commandWithPrevious) {
			throw new NotImplementedException();
		}
	}
}
