package de.ozgcloud.archive.kommentar;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.MockedStatic;

import de.ozgcloud.archive.common.UUIDConverter;
import de.ozgcloud.archive.common.VersionTypeBuilder;
import de.ozgcloud.archive.file.OzgFile;
import de.ozgcloud.archive.file.OzgFileTestFactory;
import de.xoev.xdomea.AnlageDokumentType;
import de.xoev.xdomea.DateiformatCodeType;
import de.xoev.xdomea.FormatType;
import de.xoev.xdomea.IdentifikationObjektType;
import de.xoev.xdomea.VersionType;

class DokumentTypeBuilderITCase {

	private static final String VERSION_NUMMER = "1";
	private static final String TYP = "Notiz";

	private static final int IDX_ANLAGE_PDF = 0;
	private static final int IDX_ANLAGE_JPEG = 1;
	private static final int IDX_ANLAGE_JPG = 2;

	private final Kommentar kommentar = KommentarTestFactory.create();
	private final OzgFile attachment1 = OzgFileTestFactory.createBuilder().contentType("application/pdf").build();
	private final OzgFile attachment2 = OzgFileTestFactory.createBuilder().contentType("image/jpeg").name("Testfile.jpeg").build();
	private final OzgFile attachment3 = OzgFileTestFactory.createBuilder().contentType("image/jpeg").name("Testfile.jpg").build();
	private final List<OzgFile> attachments = List.of(attachment1, attachment2, attachment3);
	private final DokumentTypeBuilder builder = DokumentTypeBuilder.builder().withKommentar(kommentar).withKommentarAttachments(attachments);

	@Nested
	class TestBuild {
		private MockedStatic<UUIDConverter> uuidConverterMockedStatic;
		private static final String KOMMENTAR_UUID = "correct kommentar ID";
		private static final String ATTACHMENT1_UUID = "correct attachment ID";

		@BeforeEach
		void setUp() {
			uuidConverterMockedStatic = mockStatic(UUIDConverter.class);
			uuidConverterMockedStatic.when(() -> UUIDConverter.fromObjectId(KommentarTestFactory.ID))
					.thenReturn(KOMMENTAR_UUID);
			uuidConverterMockedStatic.when(() -> UUIDConverter.fromObjectId(attachment1.getId().toString()))
					.thenReturn(ATTACHMENT1_UUID);
		}

		@AfterEach
		void tearDown() {
			uuidConverterMockedStatic.close();
		}

		@Test
		void shouldHaveTyp() {
			var dokumentType = builder.build();

			assertThat(dokumentType.getTyp()).isEqualTo(TYP);
		}

		@Test
		void shouldHaveIdentifikation() {
			var expectedIdentifikation = new IdentifikationObjektType();
			expectedIdentifikation.setID(KOMMENTAR_UUID);

			var dokumentType = builder.build();

			assertThat(dokumentType.getIdentifikation()).usingRecursiveComparison().isEqualTo(expectedIdentifikation);
		}

		@Nested
		class TestAnlage {
			@Test
			void shouldHaveIdentifikation() {
				var expectedIdentifikation = new IdentifikationObjektType();
				expectedIdentifikation.setID(ATTACHMENT1_UUID);

				var anlageType = getAnlage(IDX_ANLAGE_PDF);

				assertThat(anlageType.getIdentifikation()).usingRecursiveComparison().isEqualTo(expectedIdentifikation);
			}

			@Nested
			class TestVersion {
				@Test
				void shouldHaveNummer() {
					var version = getVersion(IDX_ANLAGE_PDF);

					assertThat(version.getNummer()).isEqualTo(VERSION_NUMMER);
				}

				private VersionType getVersion(int anlageIndex) {
					var anlage = getAnlage(anlageIndex);
					assertThat(anlage.getVersion()).hasSize(1);
					return getAnlage(anlageIndex).getVersion().getFirst();
				}

				@Nested
				class TestFormat {
					@Test
					void shouldHaveNameForPdf() {
						var expectedDateiFormat = createDateiformatCode("018");

						var format = getFormat(IDX_ANLAGE_PDF);

						assertThat(format.getName()).usingRecursiveComparison().isEqualTo(expectedDateiFormat);
					}

					@ParameterizedTest
					@MethodSource("shouldHaveNameForJpegDataProvider")
					void shouldHaveNameForJpeg(int anlageIdx, String expectedCode) {
						var expectedDateiFormat = createDateiformatCode(expectedCode);

						var format = getFormat(anlageIdx);

						assertThat(format.getName()).usingRecursiveComparison().isEqualTo(expectedDateiFormat);
					}

					private static Stream<Arguments> shouldHaveNameForJpegDataProvider() {
						return Stream.of(
								Arguments.of(IDX_ANLAGE_JPEG, "010"),
								Arguments.of(IDX_ANLAGE_JPG, "011"));
					}

					private DateiformatCodeType createDateiformatCode(String codeValue) {
						var dateiFormatCode = new DateiformatCodeType();
						dateiFormatCode.setCode(codeValue);
						dateiFormatCode.setListURI(VersionTypeBuilder.DATEI_FORMAT_LIST_URI);
						dateiFormatCode.setListVersionID(VersionTypeBuilder.LIST_VERSION_ID);
						return dateiFormatCode;
					}

					@Test
					void shouldHaveVersion() {
						var format = getFormat(IDX_ANLAGE_PDF);

						assertThat(format.getVersion()).isEqualTo(StringUtils.EMPTY);

					}

					private FormatType getFormat(int anlageIndex) {
						var version = getVersion(anlageIndex);
						assertThat(version.getFormat()).hasSize(1);
						return getVersion(anlageIndex).getFormat().getFirst();
					}

					@Nested
					class TestPrimaerDokument {
						@Test
						void shouldHaveDateiname() {
							var primaerdokument = getFormat(IDX_ANLAGE_PDF).getPrimaerdokument();

							assertThat(primaerdokument.getDateiname())
									.isEqualTo(ATTACHMENT1_UUID + "_" + attachment1.getName());

						}

						@Test
						void shouldHaveDateinameOriginal() {
							var primaerdokument = getFormat(IDX_ANLAGE_PDF).getPrimaerdokument();

							assertThat(primaerdokument.getDateinameOriginal())
									.isEqualTo(attachment1.getName());

						}
					}
				}

			}

			private AnlageDokumentType getAnlage(int index) {
				return builder.build().getAnlage().get(index);
			}
		}
	}
}
