package de.ozgcloud.archive.vorgang;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedConstruction;
import org.mockito.Spy;

import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextAttachingInterceptor;
import de.ozgcloud.apilib.common.callcontext.OzgCloudCallContextProvider;
import de.ozgcloud.archive.common.callcontext.CallContextProvider;
import de.ozgcloud.vorgang.vorgang.GrpcFindVorgangWithEingangRequest;
import de.ozgcloud.vorgang.vorgang.GrpcVorgangWithEingang;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;

class VorgangRemoteServiceTest {

	@Spy
	@InjectMocks
	private VorgangRemoteService service;

	@Mock
	private VorgangServiceBlockingStub grpcService;
	@Mock
	private VorgangWithEingangMapper vorgangWithEingangMapper;
	@Mock
	private CallContextProvider contextProvider;

	@Nested
	class TestFindVorgangWithEingang {

		private final VorgangWithEingang vorgangWithEingang = VorgangWithEingangTestFactory.create();
		private final GrpcVorgangWithEingang grpcVorgangWithEingang = GrpcFindVorgangWithEingangResponseTestFactory.GRPC_VORGANG_WITH_EINGANG;
		private final GrpcFindVorgangWithEingangRequest request = GrpcFindVorgangWithEingangRequestTestFactory.create();

		@Mock
		private VorgangServiceBlockingStub grpcServiceWithInterceptor;

		@BeforeEach
		void mockReturnValue() {
			doReturn(request).when(service).buildFindVorgangWithEingangRequest(VorgangWithEingangTestFactory.ID);
			doReturn(grpcServiceWithInterceptor).when(service).getGrpcServiceWithInterceptor();
			when(grpcServiceWithInterceptor.findVorgangWithEingang(request))
					.thenReturn(GrpcFindVorgangWithEingangResponseTestFactory.createVorgangWithEingangResponse());
			when(vorgangWithEingangMapper.fromGrpc(grpcVorgangWithEingang)).thenReturn(vorgangWithEingang);
		}

		@Test
		void shouldGetGrpcServiceWithInterceptor() {
			callService();

			verify(service).getGrpcServiceWithInterceptor();
		}

		@Test
		void shouldCallBuildRequest() {
			callService();

			verify(service).buildFindVorgangWithEingangRequest(VorgangWithEingangTestFactory.ID);
		}

		@Test
		void shouldCallGrpcService() {
			callService();

			verify(grpcServiceWithInterceptor).findVorgangWithEingang(request);
		}

		@Test
		void shouldCallMapper() {
			callService();

			verify(vorgangWithEingangMapper).fromGrpc(grpcVorgangWithEingang);
		}

		@Test
		void shouldReturn() {
			var result = callService();

			assertThat(result).isEqualTo(vorgangWithEingang);
		}

		private VorgangWithEingang callService() {
			return service.findVorgangWithEingang(VorgangWithEingangTestFactory.ID);
		}

	}

	@Nested
	class TestBuildRequest {

		@Test
		void shouldReturnRequest() {
			var request = service.buildFindVorgangWithEingangRequest(VorgangWithEingangTestFactory.ID);

			assertThat(request).usingRecursiveComparison().isEqualTo(GrpcFindVorgangWithEingangRequestTestFactory.create());
		}
	}

	@Nested
	class TestGetGrpcServiceWithInterceptor {

		private MockedConstruction<OzgCloudCallContextAttachingInterceptor> mockedConstructionInterceptor;
		private OzgCloudCallContextProvider argumentContextProvider;
		@Mock
		private OzgCloudCallContextAttachingInterceptor interceptor;

		@BeforeEach
		void setUpMocks() {
			mockedConstructionInterceptor = mockConstruction(OzgCloudCallContextAttachingInterceptor.class, (mock, context) -> {
				argumentContextProvider = (OzgCloudCallContextProvider) context.arguments().get(0);
				interceptor = mock;
			});
		}

		@AfterEach
		void cleanUp() {
			mockedConstructionInterceptor.close();
		}

		@Test
		void shoudCreateOzgCloudCallContextAttachingInterceptor() {
			callService();

			assertThat(mockedConstructionInterceptor.constructed()).hasSize(1);
		}

		@Test
		void shoudCreateOzgCloudCallContextAttachingInterceptorWithContextProvider() {
			callService();

			assertThat(argumentContextProvider).isEqualTo(contextProvider);
		}

		@Test
		void shouldSetInterceptor() {
			callService();

			verify(grpcService).withInterceptors(interceptor);
		}

		@Test
		void shouldReturnServiceStub() {
			when(grpcService.withInterceptors(any(OzgCloudCallContextAttachingInterceptor.class))).thenReturn(grpcService);

			var returnedServiceStub = callService();

			assertThat(returnedServiceStub).isEqualTo(grpcService);
		}

		private VorgangServiceBlockingStub callService() {
			return service.getGrpcServiceWithInterceptor();
		}
	}

}